# encoding: utf-8

from __future__ import unicode_literals
import numpy as np
import pandas as pd

from jaqs.data import DataView
from jaqs.data import RemoteDataService
from jaqs.research import SignalDigger
import jaqs.util as jutil

from config_path import DATA_CONFIG_PATH
data_config = jutil.read_json(DATA_CONFIG_PATH)

dataview_folder = '../output/prepared/test_signal'


def test_save_dataview():
    ds = RemoteDataService()
    ds.init_from_config(data_config)
    dv = DataView()
    
    props = {'start_date': 20170301, 'end_date': 20171001, 'universe': '000300.SH',
             'fields': 'volume,turnover,float_mv,pb,total_mv',
             'freq': 1}
    
    dv.init_from_config(props, ds)
    dv.prepare_data()
    
    dv.save_dataview(dataview_folder)


def test_analyze_signal():
    # --------------------------------------------------------------------------------
    # Step.1 load dataview
    dv = DataView()
    dv.load_dataview(dataview_folder)

    # --------------------------------------------------------------------------------
    # Step.2 calculate mask (to mask those ill data points)
    trade_status = dv.get_ts('trade_status')
    mask_sus = trade_status == u'停牌'

    df_index_member = dv.get_ts('index_member')
    mask_index_member = ~(df_index_member > 0)

    dv.add_formula('limit_reached', 'Abs((open - Delay(close, 1)) / Delay(close, 1)) > 0.095', is_quarterly=False)
    df_limit_reached = dv.get_ts('limit_reached')
    mask_limit_reached = df_limit_reached > 0

    mask_all = np.logical_or(mask_sus, np.logical_or(mask_index_member, mask_limit_reached))

    # --------------------------------------------------------------------------------
    # Step.3 get signal, benchmark and price data
    dv.add_formula('divert', '- Correlation(vwap_adj, volume, 10)', is_quarterly=False)
    
    signal = dv.get_ts('divert').shift(1, axis=0)  # avoid look-ahead bias
    price = dv.get_ts('close_adj')
    price_bench = dv.data_benchmark

    # Step.4 analyze!
    my_period = 5
    obj = SignalDigger(output_folder='../output/test_signal', output_format='pdf')
    obj.process_signal_before_analysis(signal, price=price,
                                       mask=mask_all,
                                       n_quantiles=5, period=my_period,
                                       benchmark_price=price_bench,
                                       )
    res = obj.create_full_report()
    

if __name__ == "__main__":
    test_save_dataview()
    test_analyze_signal()



# encoding: utf-8
from jaqs.trade.analyze.report import Report
import jaqs.util as jutil


def test_output():
    static_folder = jutil.join_relative_path('trade/analyze/static')

    r = Report({'mytitle': 'Test Title', 'mytable': 'Hello World!'},
               source_dir=static_folder,
               template_fn='test_template.html',
               out_folder='../output')
    r.generate_html()
    r.output_html()
    r.output_pdf()


if __name__ == "__main__":
    test_output()



# encoding: utf-8

from __future__ import absolute_import, print_function, unicode_literals, division
import datetime

import numpy as np
import pandas as pd

from jaqs.trade.event import EventEngine, Event, EVENT_TYPE
from jaqs.data.basic import Quote
import jaqs.util as jutil
from functools import reduce


class AlphaLiveTradeInstance(object):
    """
    Backtest alpha strategy using DataView.
    
    Attributes
    ----------
    last_rebalance_date : int
    current_rebalance_date : int
    univ_price_dic : dict
        Prices of symbols at current_date

    """
    def __init__(self):
        super(AlphaLiveTradeInstance, self).__init__()
        
        self.ctx = None
        
        self.last_date = 0
        self.last_rebalance_date = 0
        self.current_rebalance_date = 0
        
        self.univ_price_dic = {}

    def init_from_config(self, props):
        """
        
        Parameters
        ----------
        props : dict

        """
        self.props = props
    
        if 'symbol' in props:
            self.ctx.init_universe(props['symbol'])
        elif hasattr(self.ctx, 'dataview'):
            self.ctx.init_universe(self.ctx.dataview.symbol)
        else:
            raise ValueError("No dataview, no symbol either.")
    
        for obj in ['data_api', 'trade_api', 'pm', 'strategy']:
            obj = getattr(self.ctx, obj)
            if obj is not None:
                obj.init_from_config(props)
        
        self.ctx.strategy.initialize()
    
    def re_balance_plan_before_open(self):
        """
        Do portfolio re-balance before market open (not knowing suspensions) only calculate weights.
        For now, we stick to the same close price when calculate market value and do re-balance.
        
        Parameters
        ----------

        """
        # Step.1 set weights of those non-index-members to zero
        # only filter index members when universe is defined
        universe_list = self.ctx.universe
        if self.ctx.dataview.universe:
            
            # we are able to know information about index component before market open
            #col = 'index_member'
            #df_is_member = self.ctx.dataview.get_snapshot(self.ctx.trade_date, fields=col)
            #df_is_member = df_is_member.fillna(0).astype(bool)
            #dic_index_member = df_is_member.loc[:, col].to_dict()
            ser_is_member = self.ctx.dataview.get_ts('index_member').iloc[-1, :]
            dic_index_member = ser_is_member.to_dict()
            universe_list = [symbol for symbol, value in dic_index_member.items() if value]
        
        # Step.2 filter out those not listed or already de-listed
        # we are able to know information about list/delist before market open
        df_inst = self.ctx.dataview.data_inst
        mask = np.logical_and(self.ctx.trade_date > df_inst['list_date'],
                              self.ctx.trade_date < df_inst['delist_date'])
        listing_symbols = df_inst.loc[mask, :].index.values
        universe_list = [s for s in universe_list if s in listing_symbols]
        
        # step.3 construct portfolio using models
        self.ctx.strategy.portfolio_construction(universe_list)

    def get_suspensions(self):
        # trade_status = self.ctx.dataview.get_snapshot(self.ctx.trade_date, fields='trade_status')
        # trade_status = trade_status.loc[:, 'trade_status']
        # trade_status: {'N', 'XD', 'XR', 'DR', 'JiaoYi', 'TingPai', NUll (before 2003)}
        # mask_sus = trade_status == u'停牌'
        res = [k for k, v in self.univ_price_dic.items() if v['volume'] == 0]
        return res

    def get_limit_reaches(self):
        res = [k for k, v in self.univ_price_dic.items() if (v['last'] >= v['limit_up'] or v['last'] <= v['limit_down'])]
        return res
    
    def _get_latest_prices(self):
        symbols_str = ','.join(self.ctx.universe)
        cols = ','.join(['last', 'open', 'high', 'low', 'limit_up', 'limit_down', 'askprice1', 'bidprice1', 'volume', 'turnover'])
        res, _ = self.ctx.data_api.quote(symbol=symbols_str, fields=cols)
        res = res.to_dict(orient='index')
        return res
    
    def re_balance_plan_after_open(self):
        """
        Do portfolio re-balance after market open.
        With suspensions known, we re-calculate weights and generate orders.
        
        Notes
        -----
        Price here must not be adjusted.

        """
        self.univ_price_dic = self._get_latest_prices()
        prices_dic = {k: v['last'] for k, v in self.univ_price_dic.items()}
        # suspensions & limit_reaches: list of str
        suspensions = self.get_suspensions()
        limit_reaches = self.get_limit_reaches()
        all_list = reduce(lambda s1, s2: s1.union(s2), [set(suspensions), set(limit_reaches)])
        
        # step1. weights of those suspended and limit will be remove, and weights of others will be re-normalized
        self.ctx.strategy.re_weight_suspension(all_list)
        
        # step2. calculate market value and cash
        # market value does not include those suspended
        market_value_float, market_value_frozen = self.ctx.pm.market_value(prices_dic, all_list)
        cash_available = self.ctx.strategy.cash + market_value_float
        
        cash_to_use = cash_available * self.ctx.strategy.position_ratio
        cash_unuse = cash_available - cash_to_use
        
        # step3. generate target positions
        # position of those suspended will remain the same (will not be traded)
        goals, cash_remain = self.ctx.strategy.generate_weights_order(self.ctx.strategy.weights,
                                                                      cash_to_use, prices_dic,
                                                                      suspensions=all_list)
        
        self.ctx.strategy.goal_positions = self._to_valide_goals(goals)
        self.ctx.strategy.cash = cash_remain + cash_unuse
        # self.liquidate_all()
        
        self.ctx.strategy.on_after_rebalance(cash_available + market_value_frozen)
    
    def _to_valide_goals(self, goals_raw):
        univ, msg = self.ctx.trade_api.query_universe()
        univ.loc[:, 'size'] = 0.0
        univ.loc[:, 'ref_price'] = 0.0
        univ = univ.set_index('security').sort_index(axis=0)
        univ = univ[['size', 'ref_price']]
        for d in goals_raw:
            symbol = d['symbol']
            size = d['size']
            # TODO: better method needed
            if symbol in univ.index and symbol in self.univ_price_dic:
                univ.loc[symbol, 'ref_price'] = self.univ_price_dic[symbol]['last']
                if size > 0:
                    univ.loc[symbol, 'size'] = size
        goals_valid = list(univ.reset_index().to_dict(orient='index').values())
        return goals_valid
    
    @staticmethod
    def _get_current_date():
        now_utc = datetime.datetime.utcnow()
        now_utc8 = now_utc + datetime.timedelta(hours=8)
        now_date = jutil.convert_datetime_to_int(now_utc8)
        return now_date
    
    def run_alpha(self):
        tapi = self.ctx.trade_api
        
        self.ctx.trade_date = self._get_current_date()
        self.last_date = self._get_last_trade_date(self.ctx.trade_date)
        print("\n=======Re-balance starts, trade_date = {}".format(self.ctx.trade_date))
        
        # Step1.
        # we do not need adjust
        
        # Step2.
        # plan re-balance before market open of the re-balance day:
        self.ctx.snapshot = self.ctx.dataview.get_snapshot(self.last_date)
        # get index memebers, get signals, generate weights
        self.re_balance_plan_before_open()
        
        # Step3.
        # do re-balance on the re-balance day
        # get suspensions, get up/down limits, generate goal positions and send orders.
        self.re_balance_plan_after_open()
        # self.ctx.strategy.send_bullets()
        
        print("Re-balance done.")
    
    def _is_trade_date(self, date):
        if self.ctx.dataview is not None:
            return date in self.ctx.dataview.dates
        else:
            return self.ctx.data_api.is_trade_date(date)
    
    def _get_next_trade_date(self, date):
        if self.ctx.dataview is not None:
            dates = self.ctx.dataview.dates
            mask = dates > date
            return dates[mask][0]
        else:
            return self.ctx.data_api.query_next_trade_date(date)
    
    def _get_last_trade_date(self, date):
        if self.ctx.dataview is not None:
            dates = self.ctx.dataview.dates
            mask = dates < date
            return dates[mask][-1]
        else:
            return self.ctx.data_api.query_last_trade_date(date)
    
    '''
    def on_new_day(self, date):
        self.ctx.strategy.initialize()
        self.ctx.snapshot = self.ctx.dataview.get_snapshot(date)
        self.univ_price_dic = self.ctx.snapshot.loc[:, ['close', 'vwap', 'open', 'high', 'low']].to_dict(orient='index')
    
    '''
    def show_position_info(self):
        pm = self.ctx.pm
        
        prices = {k: v['open'] for k, v in self.univ_price_dic.items()}
        market_value_float, market_value_frozen = pm.market_value(prices)
        for symbol in pm.holding_securities:
            p = prices[symbol]
            size = pm.get_position(symbol).current_size
            print("{}  {:.2e}   {:.1f}@{:.2f}".format(symbol, p*size*100, p, size))
        print("float {:.2e}, frozen {:.2e}".format(market_value_float, market_value_frozen))


class EventLiveTradeInstance(EventEngine):
    """
    Attributes
    ----------
    start_date : int
    end_date : int
    
    """
    def __init__(self):
        super(EventLiveTradeInstance, self).__init__()
        
        self.start_date = 0
        self.end_date = 0
        
        self.props = None
        
        self.ctx = None
    
    def init_from_config(self, props):
        """
        
        Parameters
        ----------
        props : dict

        """
        for name in ['start_date', 'end_date']:
            if name not in props:
                pass
                # raise ValueError("{} must be provided in props.".format(name))
        
        self.props = props
        self.start_date = props.get("start_date")
        self.end_date = props.get("end_date")

        for obj in ['data_api', 'trade_api', 'pm', 'strategy']:
            if hasattr(self.ctx, obj):
                getattr(self.ctx, obj).init_from_config(props)

    def register_context(self, context=None):
        self.ctx = context

    # -------------------------------------------------------------------------------------------
    # Run
    def run(self):
        """
        Listen to certain events and run the EventEngine.
        Events include:
            1. market_data are from DataService
            2. trades & orders indications are from TradeApi.
            3. etc.

        """

        self.register(EVENT_TYPE.MARKET_DATA, self.on_bar)
        
        self.register(EVENT_TYPE.TASK_STATUS_IND, self.on_task_status)
        self.register(EVENT_TYPE.ORDER_RSP, self.on_order_rsp)
        self.register(EVENT_TYPE.TASK_RSP, self.on_task_rsp)
        self.register(EVENT_TYPE.TRADE_IND, self.on_trade)
        self.register(EVENT_TYPE.ORDER_STATUS_IND, self.on_order_status)
        
        self.start(timer=False)
    
    def on_bar(self, event):
        quote_dic = event.dic['quote']
        quote = Quote.create_from_dict(quote_dic)
        self.ctx.strategy.on_tick(quote)
    
    def on_order_rsp(self, event):
        rsp = event.dic['rsp']
        self.ctx.strategy.on_order_rsp(rsp)

    def on_task_rsp(self, event):
        rsp = event.dic['rsp']
        self.ctx.strategy.on_task_rsp(rsp)
    
    def on_trade(self, event):
        ind = event.dic['ind']
        self.ctx.strategy.on_trade(ind)

    def on_order_status(self, event):
        ind = event.dic['ind']
        self.ctx.strategy.on_order_status(ind)

    def on_task_status(self, event):
        ind = event.dic['ind']
        self.ctx.strategy.on_task_status(ind)
    
    # ---------------------------------------------------------
    # Save Results
    def save_results(self, folder_path='.'):
        import os
        import pandas as pd
        folder_path = os.path.abspath(folder_path)
    
        trades = self.ctx.pm.trades
    
        type_map = {'task_id': str,
                    'entrust_no': str,
                    'entrust_action': str,
                    'symbol': str,
                    'fill_price': float,
                    'fill_size': float,
                    'fill_date': np.integer,
                    'fill_time': np.integer,
                    'fill_no': str,
                    'commission': float}
        # keys = trades[0].__dict__.keys()
        ser_list = dict()
        for key in type_map.keys():
            v = [t.__getattribute__(key) for t in trades]
            ser = pd.Series(data=v, index=None, dtype=type_map[key], name=key)
            ser_list[key] = ser
        df_trades = pd.DataFrame(ser_list)
        df_trades.index.name = 'index'
    
        trades_fn = os.path.join(folder_path, 'trades.csv')
        configs_fn = os.path.join(folder_path, 'configs.json')
        jutil.create_dir(trades_fn)
    
        df_trades.to_csv(trades_fn)
        jutil.save_json(self.props, configs_fn)
    
        print ("Backtest results has been successfully saved to:\n" + folder_path)



# encoding:utf-8


class Position(object):
    """
    Basic position class.

    Attributes
    ----------
    symbol : str
        List of securities.
    side : str
        ("Long", "Short"). Positions of different sides will not be merged.
    cost_price : float
        Average cost price of current net position.
    close_pnl : float
    float_pnl : float
    trading_pnl : float
    holding_pnl : float
    init_size : int
        Position at the start of the day.
    enable_size : int
        Position that can be closed.
    frozen_size : int
        Positions to be closed.
    want_size : int
        Positions to be opened.
    pre_size : int
        Last day's position.
    today_size : int
        Today's position.
    current_size : int
        Current position.

    Methods
    -------

    """
    '''
    __slots__ = ['symbol', 'side', 'cost_price',
                 'close_pnl', 'float_pnl', 'trading_pnl', 'holding_pnl',
                 'enable_size', 'frozen_size', 'want_size',
                 'today_size', 'pre_size', 'current_size', 'init_size',
                 'commission']
    '''
    
    def __init__(self, symbol=""):
        self.symbol = symbol
        
        self.side = ""
        self.cost_price = 0.0
        
        self.close_pnl = 0.0
        self.float_pnl = 0.0
        self.trading_pnl = 0.0
        self.holding_pnl = 0.0
        
        self.enable_size = 0
        self.frozen_size = 0
        self.want_size = 0
        
        self.today_size = 0
        self.pre_size = 0
        self.current_size = 0
        self.init_size = 0
        
        self.commission = 0.0

    def __repr__(self):
        return "{0.side:7s} {0.current_size:5d} of {0.symbol:10s}".format(self)

    def __str__(self):
        return self.__repr__()

    @classmethod
    def create_from_df(cls, df):
        bar_list = []
        for _, row in df.iterrows():
            dic = row.to_dict()
            bar = cls.create_from_dict(dic)
            bar_list.append(bar)
        return bar_list

    @classmethod
    def create_from_dict(cls, dic):
        bar = cls()
        bar.__dict__.update(dic)
        return bar


class GoalPosition(object):
    """
    Used in goal_portfolio function to generate orders.

    Attributes
    ----------
    symbol : str
    ref_price : float
        Reference price, used by risk management, not by order.
    size : int
        Target position size.
    urgency : int
        The urgency to adjust position, used to determine trading algorithm.

    """
    
    def __init__(self):
        self.symbol = ""
        self.ref_price = 0.0
        self.size = 0
        self.urgency = 0
    
    def __repr__(self):
        return "{0.size:5d} of {0.symbol:10s}".format(self)

    def __str__(self):
        return self.__repr__()



# encoding: utf-8

from __future__ import print_function
from jaqs.trade import model
import jaqs.util as jutil
import random


def test_context():
    r = random.random()
    path = '../../output/tests/storage{:.6f}.pic'.format(r)
    context = model.Context()
    context.load_store(path)
    assert len(context.storage) == 0
    context.storage['me'] = 1.0
    context.save_store(path)
    
    context = model.Context()
    context.load_store(path)
    assert context.storage['me'] == 1.0


if __name__ == "__main__":
    import time
    t_start = time.time()
    
    g = globals()
    g = {k: v for k, v in g.items() if k.startswith('test_') and callable(v)}
    
    for test_name, test_func in g.items():
        print("\n==========\nTesting {:s}...".format(test_name))
        test_func()
    print("Test Complete.")
    
    t3 = time.time() - t_start
    print("\n\n\nTime lapsed in total: {:.1f}".format(t3))



# encoding: utf-8

from __future__ import print_function
import os
from collections import OrderedDict

import numpy as np
import pandas as pd

from . import performance as pfm
from . import plotting

import jaqs.util as jutil
from jaqs.trade import common


class SignalDigger(object):
    """
    
    Attributes
    ----------
    signal_data : pd.DataFrame - MultiIndex
        Index is pd.MultiIndex ['trade_date', 'symbol'], columns = ['signal', 'return', 'quantile']
    period : int
        Horizon used to calculate return.
    n_quantiles : int
    output_format : str
    output_folder : str
        
    """
    def __init__(self, output_folder=".", output_format='pdf'):
        self.output_format = output_format
        self.output_folder = os.path.abspath(output_folder)
        
        self.signal_data = None
        self.period = None
        self.n_quantiles = 5
        self.benchmark_ret = None
        
        self.returns_report_data = dict()
        self.ic_report_data = dict()
        self.fig_data = dict()
        self.fig_objs = dict()

    def process_signal_before_analysis(self,
                                       signal, price=None, ret=None, benchmark_price=None,
                                       period=5, n_quantiles=5,
                                       mask=None,
                                       forward=False):
        """
        Prepare for signal analysis.

        Parameters
        ----------
        signal : pd.DataFrame
            Index is date, columns are stocks.
        price : pd.DataFrame
            Index is date, columns are stocks.
        ret : pd.DataFrame
            Index is date, columns are stocks.
        benchmark_price : pd.DataFrame or pd.Series or None
            Price of benchmark.
        mask : pd.DataFrame
            Data cells that should NOT be used.
        n_quantiles : int
        period : int
            periods to compute forward returns on.

        Returns
        -------
        res : pd.DataFrame
            Index is pd.MultiIndex ['trade_date', 'symbol'], columns = ['signal', 'return', 'quantile']
            
        """
        """
        Deal with suspensions:
            If the period of calculating return is d (from T to T+d), then
            we do not use signal values of those suspended on T,
            we do not calculate return for those suspended on T+d.
        """
        # ----------------------------------------------------------------------
        # parameter validation
        if price is None and ret is None:
            raise ValueError("One of price / ret must be provided.")
        if price is not None and ret is not None:
            raise ValueError("Only one of price / ret should be provided.")
        if ret is not None and benchmark_price is not None:
            raise ValueError("You choose 'return' mode but benchmark_price is given.")
        if not (n_quantiles > 0 and isinstance(n_quantiles, int)):
            raise ValueError("n_quantiles must be a positive integer. Input is: {}".format(n_quantiles))
        
        # ensure inputs are aligned
        data = price if price is not None else ret
        assert np.all(signal.index == data.index)
        assert np.all(signal.columns == data.columns)
        if mask is not None:
            assert np.all(signal.index == mask.index)
            assert np.all(signal.columns == mask.columns)
            mask = jutil.fillinf(mask)
            mask = mask.astype(int).fillna(0).astype(bool)  # dtype of mask could be float. So we need to convert.
        else:
            mask = pd.DataFrame(index=signal.index, columns=signal.columns, data=False)
        signal = jutil.fillinf(signal)
        data = jutil.fillinf(data)

        # ----------------------------------------------------------------------
        # save data
        self.n_quantiles = n_quantiles
        self.period = period

        # ----------------------------------------------------------------------
        # Get dependent variables
        if price is not None:
            df_ret = pfm.price2ret(price, period=self.period, axis=0)
            if benchmark_price is not None:
                benchmark_price = benchmark_price.loc[signal.index]
                bench_ret = pfm.price2ret(benchmark_price, self.period, axis=0)
                self.benchmark_ret = bench_ret
                residual_ret = df_ret.sub(bench_ret.values.flatten(), axis=0)
            else:
                residual_ret = df_ret
        else:
            residual_ret = ret
        
        # Get independent varibale
        signal = signal.shift(1)  # avoid forward-looking bias

        # forward or not
        if forward:
            # point-in-time signal and forward return
            residual_ret = residual_ret.shift(-self.period)
        else:
            # past signal and point-in-time return
            signal = signal.shift(self.period)

        # ----------------------------------------------------------------------
        # get masks
        # mask_prices = data.isnull()
        # Because we use FORWARD return, if one day's price is broken, the day that is <period> days ago is also broken.
        # mask_prices = np.logical_or(mask_prices, mask_prices.shift(self.period))
        mask_price_return = residual_ret.isnull()
        mask_signal = signal.isnull()

        mask = np.logical_or(mask_signal, mask_price_return)
        # mask = np.logical_or(mask, mask_signal)

        # if price is not None:
        #     mask_forward = np.logical_or(mask, mask.shift(self.period).fillna(True))
        #     mask = np.logical_or(mask, mask_forward)

        # ----------------------------------------------------------------------
        # calculate quantile
        signal_masked = signal.copy()
        signal_masked = signal_masked[~mask]
        if n_quantiles == 1:
            df_quantile = signal_masked.copy()
            df_quantile.loc[:, :] = 1.0
        else:
            df_quantile = jutil.to_quantile(signal_masked, n_quantiles=n_quantiles)

        # ----------------------------------------------------------------------
        # stack
        def stack_td_symbol(df):
            df = pd.DataFrame(df.stack(dropna=False))  # do not dropna
            df.index.names = ['trade_date', 'symbol']
            df.sort_index(axis=0, level=['trade_date', 'symbol'], inplace=True)
            return df
        
        mask = stack_td_symbol(mask)
        df_quantile = stack_td_symbol(df_quantile)
        residual_ret = stack_td_symbol(residual_ret)

        # ----------------------------------------------------------------------
        # concat signal value
        res = stack_td_symbol(signal)
        res.columns = ['signal']
        res['return'] = residual_ret
        res['quantile'] = df_quantile
        res = res.loc[~(mask.iloc[:, 0]), :]
        
        print("Nan Data Count (should be zero) : {:d};  " \
              "Percentage of effective data: {:.0f}%".format(res.isnull().sum(axis=0).sum(),
                                                             len(res) * 100. / signal.size))
        res = res.astype({'signal': float, 'return': float, 'quantile': int})
        self.signal_data = res
    
    def show_fig(self, fig, file_name):
        """
        Save fig object to self.output_folder/filename.
        
        Parameters
        ----------
        fig : matplotlib.figure.Figure
        file_name : str

        """
        
        self.fig_objs[file_name] = fig
        
        if self.output_format in ['pdf', 'png', 'jpg']:
            fp = os.path.join(self.output_folder, '.'.join([file_name, self.output_format]))
            jutil.create_dir(fp)
            fig.savefig(fp)
            print("Figure saved: {}".format(fp))
        elif self.output_format == 'base64':
            fig_b64 = jutil.fig2base64(fig, 'png')
            self.fig_data[file_name] = fig_b64
            print("Base64 data of figure {} will be stored in dictionary.".format(file_name))
        elif self.output_format == 'plot':
            fig.show()
        else:
            raise NotImplementedError("output_format = {}".format(self.output_format))
    
    @plotting.customize
    def create_returns_report(self):
        """
        Creates a tear sheet for returns analysis of a signal.

        """
        n_quantiles = self.signal_data['quantile'].max()
        
        # ----------------------------------------------------------------------------------
        # Daily Signal Return Time Series
        # Use regression or weighted average to calculate.
        period_wise_long_ret =\
            pfm.calc_period_wise_weighted_signal_return(self.signal_data, weight_method='long_only')
        period_wise_short_ret = \
            pfm.calc_period_wise_weighted_signal_return(self.signal_data, weight_method='short_only')
        cum_long_ret = pfm.period_wise_ret_to_cum(period_wise_long_ret, period=self.period, compound=False)
        cum_short_ret = pfm.period_wise_ret_to_cum(period_wise_short_ret, period=self.period, compound=False)
        # period_wise_ret_by_regression = perf.regress_period_wise_signal_return(signal_data)
        # period_wise_ls_signal_ret = \
        #     pfm.calc_period_wise_weighted_signal_return(signal_data, weight_method='long_short')
        # daily_ls_signal_ret = pfm.period2daily(period_wise_ls_signal_ret, period=period)
        # ls_signal_ret_cum = pfm.daily_ret_to_cum(daily_ls_signal_ret)

        # ----------------------------------------------------------------------------------
        # Period-wise Quantile Return Time Series
        # We calculate quantile return using equal weight or market value weight.
        # Quantile is already obtained according to signal values.
        
        # quantile return
        period_wise_quantile_ret_stats = pfm.calc_quantile_return_mean_std(self.signal_data, time_series=True)
        cum_quantile_ret = pd.concat({k: pfm.period_wise_ret_to_cum(v['mean'], period=self.period, compound=False)
                                      for k, v in period_wise_quantile_ret_stats.items()},
                                     axis=1)
        
        # top quantile minus bottom quantile return
        period_wise_tmb_ret = pfm.calc_return_diff_mean_std(period_wise_quantile_ret_stats[n_quantiles],
                                                            period_wise_quantile_ret_stats[1])
        cum_tmb_ret = pfm.period_wise_ret_to_cum(period_wise_tmb_ret['mean_diff'], period=self.period, compound=False)

        # ----------------------------------------------------------------------------------
        # Alpha and Beta
        # Calculate using regression.
        '''
        weighted_portfolio_alpha_beta
        tmb_alpha_beta =
        '''
        
        # start plotting
        if self.output_format:
            vertical_sections = 6
            gf = plotting.GridFigure(rows=vertical_sections, cols=1)
            gf.fig.suptitle("Returns Tear Sheet\n\n(no compound)\n (period length = {:d} days)".format(self.period))
    
            plotting.plot_quantile_returns_ts(period_wise_quantile_ret_stats,
                                              ax=gf.next_row())

            plotting.plot_cumulative_returns_by_quantile(cum_quantile_ret,
                                                         ax=gf.next_row())

            plotting.plot_cumulative_return(cum_long_ret,
                                            title="Signal Weighted Long Only Portfolio Cumulative Return",
                                            ax=gf.next_row())
            
            plotting.plot_cumulative_return(cum_short_ret,
                                            title="Signal Weighted Short Only Portfolio Cumulative Return",
                                            ax=gf.next_row())

            plotting.plot_mean_quantile_returns_spread_time_series(period_wise_tmb_ret, self.period,
                                                                   bandwidth=0.5,
                                                                   ax=gf.next_row())
            
            plotting.plot_cumulative_return(cum_tmb_ret,
                                            title="Top Minus Bottom (long top, short bottom)"
                                                  "Portfolio Cumulative Return",
                                            ax=gf.next_row())

            self.show_fig(gf.fig, 'returns_report')
        
        self.returns_report_data = {'period_wise_quantile_ret': period_wise_quantile_ret_stats,
                                    'cum_quantile_ret': cum_quantile_ret,
                                    'cum_long_ret': cum_long_ret,
                                    'cum_short_ret': cum_short_ret,
                                    'period_wise_tmb_ret': period_wise_tmb_ret,
                                    'cum_tmb_ret': cum_tmb_ret}

    @plotting.customize
    def create_information_report(self):
        """
        Creates a tear sheet for information analysis of a signal.
        
        """
        ic = pfm.calc_signal_ic(self.signal_data)
        ic.index = pd.to_datetime(ic.index, format="%Y%m%d")
        monthly_ic = pfm.mean_information_coefficient(ic, "M")

        if self.output_format:
            ic_summary_table = pfm.calc_ic_stats_table(ic)
            plotting.plot_information_table(ic_summary_table)
            
            columns_wide = 2
            fr_cols = len(ic.columns)
            rows_when_wide = (((fr_cols - 1) // columns_wide) + 1)
            vertical_sections = fr_cols + 3 * rows_when_wide + 2 * fr_cols
            gf = plotting.GridFigure(rows=vertical_sections, cols=columns_wide)
            gf.fig.suptitle("Information Coefficient Report\n\n(period length = {:d} days)"
                            "\ndaily IC = rank_corr(period-wise forward return, signal value)".format(self.period))

            plotting.plot_ic_ts(ic, self.period, ax=gf.next_row())
            plotting.plot_ic_hist(ic, self.period, ax=gf.next_row())
            # plotting.plot_ic_qq(ic, ax=ax_ic_hqq[1::2])

            plotting.plot_monthly_ic_heatmap(monthly_ic, period=self.period, ax=gf.next_row())
        
            self.show_fig(gf.fig, 'information_report')
        
        self.ic_report_data = {'daily_ic': ic,
                               'monthly_ic': monthly_ic}
    
    def create_binary_event_report(self, signal, price, mask, benchmark_price, periods,
                                   join_method_periods='inner', group_by=None):
        """
        
        Parameters
        ----------
        signal : pd.DataFrame
        price : pd.DataFrame
        mask : pd.DataFrame
        benchmark_price : pd.DataFrame
        periods : list of int
        join_method_periods : {'inner', 'outer'}.
            Whether to take intersection or union of data of different periods.
        group_by : {'year', 'month', None}
            Calculate various statistics within each year/month/whole sample.

        Returns
        -------
        res : dict

        """
        import scipy.stats as scst

        # Raw Data
        dic_signal_data = OrderedDict()
        for my_period in periods:
            self.process_signal_before_analysis(signal, price=price, mask=mask,
                                                n_quantiles=1, period=my_period,
                                                benchmark_price=benchmark_price,
                                                forward=True)
            dic_signal_data[my_period] = self.signal_data
        
        # Processed Data
        dic_events = OrderedDict()
        dic_all = OrderedDict()
        for period, df in dic_signal_data.items():
            ser_ret = df['return']
            ser_sig = df['signal'].astype(bool)
            events_ret = ser_ret.loc[ser_sig]
            dic_events[period] = events_ret
            dic_all[period] = ser_ret
        df_events = pd.concat(dic_events, axis=1, join=join_method_periods)
        df_all = pd.concat(dic_all, axis=1, join=join_method_periods)

        # Data Statistics
        def _calc_statistics(df):
            df_res = pd.DataFrame(index=periods,
                                  columns=['Annu. Ret.', 'Annu. Vol.',
                                           #'Annual Return (all sample)', 'Annual Volatility (all sample)',
                                           't-stat', 'p-value', 'skewness', 'kurtosis', 'occurance'],
                                  data=np.nan)
            df_res.index.name = 'Period'
            
            ser_periods = pd.Series(index=df.columns, data=df.columns.values)
            ratio = (1.0 * common.CALENDAR_CONST.TRADE_DAYS_PER_YEAR / ser_periods)
            mean = df.mean(axis=0)
            std = df.std(axis=0)
            annual_ret, annual_vol = mean * ratio, std * np.sqrt(ratio)
            
            t_stats, p_values = scst.ttest_1samp(df.values, np.zeros(df.shape[1]), axis=0)
            df_res.loc[:, 't-stat'] = t_stats
            df_res.loc[:, 'p-value'] = np.round(p_values, 5)
            df_res.loc[:, "skewness"] = scst.skew(df, axis=0)
            df_res.loc[:, "kurtosis"] = scst.kurtosis(df, axis=0)
            df_res.loc[:, 'Annu. Ret.'] = annual_ret
            df_res.loc[:, 'Annu. Vol.'] = annual_vol
            df_res.loc[:, 'occurance'] = len(df)
            # dic_res[period] = df
            return df_res

        if group_by == 'year':
            grouper_func = jutil.date_to_year
        elif group_by == 'month':
            grouper_func = jutil.date_to_month
        else:
            grouper_func = get_dummy_grouper

        idx_group = grouper_func(df_events.index.get_level_values('trade_date'))
        df_stats = df_events.groupby(idx_group).apply(_calc_statistics)
        idx_group_all = grouper_func(df_all.index.get_level_values('trade_date'))
        df_all_stats = df_all.groupby(idx_group_all).apply(_calc_statistics)
        df_all_stats = df_all_stats.loc[df_stats.index, ['Annu. Ret.', 'Annu. Vol.']]
        df_all_stats.columns = ['Annu. Ret. (all samp)', 'Annu. Vol. (all samp)']
        df_stats = pd.concat([df_stats, df_all_stats], axis=1)

        # return df_all, df_events, df_stats
        ser_signal_raw, monthly_signal, yearly_signal = calc_calendar_distribution(signal)

        # return
        # plot
        gf = plotting.GridFigure(rows=len(np.unique(idx_group)) * len(periods) + 3, cols=2, height_ratio=1.2)
        gf.fig.suptitle("Event Return Analysis (annualized)")

        plotting.plot_calendar_distribution(ser_signal_raw,
                                            monthly_signal=monthly_signal, yearly_signal=yearly_signal,
                                            ax1=gf.next_row(), ax2=gf.next_row())
        plotting.plot_event_bar(df_stats.reset_index(), x='Period', y='Annu. Ret.', hue='trade_date', ax=gf.next_row())
        # plotting.plot_event_pvalue(df_stats['p-value'], ax=gf.next_subrow())
        
        def _plot_dist(df):
            date = grouper_func(df.index.get_level_values('trade_date'))[0]
            plotting.plot_event_dist(df, group_by.title() + ' ' + str(date), axs=[gf.next_cell() for _ in periods])
        if group_by is not None:
            df_events.groupby(idx_group).apply(_plot_dist)
        else:
            plotting.plot_event_dist(df_events, "", axs=[gf.next_cell() for _ in periods])
        
        self.show_fig(gf.fig, 'event_report')

        # dic_res['df_res'] = df_res
        return df_all, df_events, df_stats
        
    @plotting.customize
    def create_full_report(self):
        """
        Creates a full tear sheet for analysis and evaluating single
        return predicting (alpha) signal.
        
        """
        # signal quantile description statistics
        qstb = calc_quantile_stats_table(self.signal_data)
        if self.output_format:
            plotting.plot_quantile_statistics_table(qstb)
            
        self.create_returns_report()
        self.create_information_report()
        # we do not do turnover analysis for now
        # self.create_turnover_report(signal_data)
        
        res = dict()
        res.update(self.returns_report_data)
        res.update(self.ic_report_data)
        res.update(self.fig_data)
        return res

    def create_single_signal_report(self, signal, price, periods, n_quantiles, mask=None, trade_condition=None):
        """
        
        Parameters
        ----------
        signal : pd.Series
        index is integer date, values are signals
        price : pd.Series
        index is integer date, values are prices
        mask : pd.Series or None, optional
        index is integer date, values are bool
        periods : list of int
        trade_condition : dict , optional
            {'cond_name1': {'col_name': str, 'hold': int, 'filter': func, 'direction': 1},
             'cond_name2': {'col_name': str, 'hold': int, 'filter': func, 'direction': -1},
            }
        
        Returns
        -------
        res : dict
        
        """
        if isinstance(signal, pd.DataFrame):
            signal = signal.iloc[:, 0]
        if isinstance(price, pd.DataFrame):
            price = price.iloc[:, 0]
            
        # calc return
        ret_l = {period: pfm.price2ret(price, period=period, axis=0) for period in periods}
        df_ret = pd.concat(ret_l, axis=1)

        # ----------------------------------------------------------------------
        # calculate quantile
        if n_quantiles == 1:
            df_quantile = signal.copy()
            df_quantile.loc[:] = 1.0
        else:
            df_quantile = jutil.to_quantile(signal, n_quantiles=n_quantiles, axis=0)

        # ----------------------------------------------------------------------
        # concat signal value
        res = pd.DataFrame(signal.shift(1))
        res.columns = ['signal']
        res['quantile'] = df_quantile
        res = pd.concat([res, df_ret], axis=1)
        res = res.dropna()

        print("Nan Data Count (should be zero) : {:d};  " \
              "Percentage of effective data: {:.0f}%".format(res.isnull().sum(axis=0).sum(),
                                                             len(res) * 100. / signal.size))
        
        # calc quantile stats
        gp = res.groupby(by='quantile')
        dic_raw = {k: v for k, v in gp}
        dic_stats = OrderedDict()
        for q, df in gp:
            df_stat = pd.DataFrame(index=['mean', 'std'], columns=df_ret.columns, data=np.nan)
            df_stat.loc['mean', :] = df.loc[:, df_ret.columns].mean(axis=0)
            df_stat.loc['std', :] = df.loc[:, df_ret.columns].std(axis=0)
            dic_stats[q] = df_stat
        
        # calculate IC
        ics = calc_various_ic(res, ret_cols=df_ret.columns)
        
        # backtest
        if trade_condition is not None:
            def sim_backtest(df, dic_of_cond):
                dic_cum_ret = dict()
                for key, dic in dic_of_cond.items():
                    col_name = dic['column']
                    func = dic['filter']
                    n_hold = dic['hold']
                    direction = dic['direction']
                    mask = df[col_name].apply(func).astype(int)
                    dic_cum_ret[key] = (df[n_hold] * mask).cumsum() * direction
                df_cumret = pd.concat(dic_cum_ret, axis=1)
                return df_cumret
            df_backtest = sim_backtest(res, trade_condition)
            
        # plot
        gf = plotting.GridFigure(rows=3, cols=1, height_ratio=1.2)
        gf.fig.suptitle("Event Return Analysis (annualized)")
        
        plotting.plot_ic_decay(ics, ax=gf.next_row())
        
        plotting.plot_quantile_return_mean_std(dic_stats, ax=gf.next_row())
        
        if trade_condition is not None:
            plotting.plot_batch_backtest(df_backtest, ax=gf.next_row())
        
        self.show_fig(gf.fig, 'single_inst.pdf')


def calc_ic(x, y, method='rank'):
    """
    Calculate IC between x and y.
    
    Parameters
    ----------
    x : np.ndarray
    y : np.ndarray
    method : {'rank', 'normal'}

    Returns
    -------
    corr : float

    """
    import scipy.stats as scst
    if method == 'rank':
        corr = scst.spearmanr(x, y)[0]
    elif method == 'normal':
        corr = np.corrcoef(x, y)[0, 1]
    else:
        raise NotImplementedError("method = {}".format(method))
    return corr


def calc_various_ic(df, ret_cols):
    res_dic = dict()
    
    # signal normal IC: signal value v.s. return
    res_dic['normal'] = [calc_ic(df['signal'], df[col], method='normal') for col in ret_cols]
    
    # signal rank IC: signal value v.s. return
    res_dic['rank'] = [calc_ic(df['signal'], df[col], method='rank') for col in ret_cols]
    
    # quantile normal IC: signal quantile v.s. return
    res_dic['normal_q'] = [calc_ic(df['quantile'], df[col], method='normal') for col in ret_cols]
    
    # quantile rank IC: signal quantile v.s. return
    res_dic['rank_q'] = [calc_ic(df['quantile'], df[col], method='rank') for col in ret_cols]
    
    res = pd.DataFrame(index=ret_cols, data=res_dic)
    return res
    

def calc_quantile_stats_table(signal_data):
    quantile_stats = signal_data.groupby('quantile').agg(['min', 'max', 'mean', 'std', 'count'])['signal']
    quantile_stats['count %'] = quantile_stats['count'] / quantile_stats['count'].sum() * 100.
    return quantile_stats


def get_dummy_grouper(ser):
    res = pd.Index(np.array(['all_sample'] * len(ser)), name=ser.name)
    return res
    
    
def calc_calendar_distribution(df_signal):
    daily_signal = df_signal.sum(axis=1)
    daily_signal = daily_signal.fillna(0).astype(int)
    idx = daily_signal.index.values
    month = jutil.date_to_month(idx)
    year = jutil.date_to_year(idx)
    
    monthly_signal = daily_signal.groupby(by=month).sum()
    yearly_signal = daily_signal.groupby(by=year).sum()
    
    monthly_signal = pd.DataFrame(monthly_signal, columns=['Time'])
    yearly_signal = pd.DataFrame(yearly_signal, columns=['Time'])
    monthly_signal.index.name = 'Month'
    yearly_signal.index.name = 'Year'
    
    return daily_signal, monthly_signal, yearly_signal



# -*- encoding: utf-8 -*-
"""
Both backtest and live trading are included.
"""

from __future__ import print_function
from __future__ import absolute_import
import time

from jaqs.data import RemoteDataService
from jaqs.trade import AlphaBacktestInstance, AlphaLiveTradeInstance

import jaqs.util as jutil
from jaqs.trade import PortfolioManager
import jaqs.trade.analyze as ana
from jaqs.trade import AlphaStrategy
from jaqs.trade import AlphaTradeApi, RealTimeTradeApi
from jaqs.trade import model
from jaqs.data import DataView

from config_path import DATA_CONFIG_PATH, TRADE_CONFIG_PATH
data_config = jutil.read_json(DATA_CONFIG_PATH)
trade_config = jutil.read_json(TRADE_CONFIG_PATH)

dataview_dir_path = '../output/wine_industry_momentum/dataview'
backtest_result_dir_path = '../output/wine_industry_momentum'

BENCHMARK = '399997.SZ'


def test_save_dataview():
    ds = RemoteDataService()
    ds.init_from_config(data_config)
    dv = DataView()

    props = {'start_date': 20170901, 'end_date': 20171129, 'universe': BENCHMARK,
             'fields': 'close,volume,sw1',
             'freq': 1}

    dv.init_from_config(props, ds)
    dv.prepare_data()

    dv.add_formula('ret', 'Return(close_adj, 20)', is_quarterly=False)
    dv.add_formula('rank_ret', 'Rank(ret)', is_quarterly=False)

    dv.save_dataview(folder_path=dataview_dir_path)


def my_selector(context, user_options=None):
    rank_ret = context.snapshot['rank_ret']

    return rank_ret >= 0.9


def test_backtest():
    dv = DataView()
    dv.load_dataview(folder_path=dataview_dir_path)
    
    props = {
        "benchmark": BENCHMARK,
        "universe": ','.join(dv.symbol),
        
        "start_date": dv.start_date,
        "end_date": dv.end_date,
        
        "period": "day",
        "days_delay": 0,
        
        "init_balance": 1e8,
        "position_ratio": 1.0,
    }
    props.update(data_config)
    props.update(trade_config)
    
    stock_selector = model.StockSelector()
    stock_selector.add_filter(name='rank_ret_top10', func=my_selector)
    
    strategy = AlphaStrategy(stock_selector=stock_selector, pc_method='equal_weight')
    pm = PortfolioManager()
    bt = AlphaBacktestInstance()
    trade_api = AlphaTradeApi()
    ds = None
    
    context = model.Context(dataview=dv, instance=bt, strategy=strategy, trade_api=trade_api, pm=pm, data_api=ds)
    stock_selector.register_context(context)
    
    bt.init_from_config(props)
    bt.run_alpha()
    
    bt.save_results(folder_path=backtest_result_dir_path)

    do_analyze()


def test_livetrade():
    dv = DataView()
    dv.load_dataview(folder_path=dataview_dir_path)
    
    props = {
        "benchmark": BENCHMARK,
        "universe": ','.join(dv.symbol),
        
        "start_date": dv.start_date,
        "end_date": dv.end_date,
        
        "period": "day",
        "days_delay": 0,
        
        "init_balance": 1e8,
        "position_ratio": 1.0,
        "strategy_no": 1044
    }
    props.update(data_config)
    props.update(trade_config)
    
    stock_selector = model.StockSelector()
    stock_selector.add_filter(name='rank_ret_top10', func=my_selector)
    
    strategy = AlphaStrategy(stock_selector=stock_selector, pc_method='equal_weight')
    pm = PortfolioManager()
    bt = AlphaLiveTradeInstance()
    trade_api = RealTimeTradeApi(props)
    ds = RemoteDataService()
    
    context = model.Context(dataview=dv, instance=bt, strategy=strategy, trade_api=trade_api, pm=pm, data_api=ds)
    stock_selector.register_context(context)
    
    bt.init_from_config(props)
    bt.run_alpha()
    
    goal_positions = strategy.goal_positions
    print(goal_positions)
    
    do_analyze()


def do_analyze():
    ta = ana.AlphaAnalyzer()
    dv = DataView()
    dv.load_dataview(folder_path=dataview_dir_path)

    ta.initialize(dataview=dv, file_folder=backtest_result_dir_path)
    
    ta.do_analyze(result_dir=backtest_result_dir_path, selected_sec=list(ta.universe)[:3], brinson_group='sw1')


if __name__ == "__main__":
    t_start = time.time()

    test_save_dataview()
    test_backtest()
    test_livetrade()

    t3 = time.time() - t_start
    print("\n\n\nTime lapsed in total: {:.1f}".format(t3))



# encoding: UTF-8

from __future__ import print_function
import time

from jaqs.trade.tradeapi import TradeApi
import pandas as pd
import jaqs.util as jutil

from config_path import TRADE_CONFIG_PATH
trade_config = jutil.read_json(TRADE_CONFIG_PATH)


def test_trade_api():
    address = trade_config.get("remote.trade.address", None)
    username = trade_config.get("remote.trade.username", None)
    password = trade_config.get("remote.trade.password", None)
    print(address)
    print(username)
    # print password
    
    if address is None or username is None or password is None:
        raise ValueError("no trade service config available!")
    
    tapi = TradeApi(address)

    # TradeApi通过回调函数方式通知用户事件。事件包括三种：订单状态、成交回报、委托任务执行状态。

    # 订单状态推送
    def on_orderstatus(order):
        print("on_orderstatus:") #, order
        for key in order:    print("%20s : %s" % (key, str(order[key])))
        print("")


    # 成交回报推送
    def on_trade(trade):
        print("on_trade:")
        for key in trade:    print("%20s : %s" % (key, str(trade[key])))
        print("")

    # 委托任务执行状态推送
    # 通常可以忽略该回调函数
    def on_taskstatus(task):
        print("on_taskstatus:")
        for key in task:    print("%20s : %s" % (key, str(task[key])))
        print("")

    tapi.set_ordstatus_callback(on_orderstatus)
    tapi.set_trade_callback(on_trade)
    tapi.set_task_callback(on_taskstatus)
    
    # 使用用户名、密码登陆， 如果成功，返回用户可用的策略帐号列表
    user_info, msg = tapi.login(username, password)
    print("msg: ", msg)
    assert check_err_msg(msg)
    print("user_info:", user_info)
    user_strats = user_info['strategies']

    # 选择使用的策略帐号
    #
    # 该函数成功后，下单、查持仓等和策略帐号有关的操作都和该策略帐号绑定。
    # 没有必要每次下单、查询都调用该函数。重复调用该函数可以选择新的策略帐号。
    #
    # 如果成功，返回(strategy_id, msg)
    # 否则返回 (0, err_msg)
    if user_strats:
        sid, msg = tapi.use_strategy(user_strats[0])
        assert check_err_msg(msg)
        print("sid: ", sid)
        time.sleep(1)

    # 查询Portfolio
    #
    # 返回当前的策略帐号的Universe中所有标的的净持仓，包括持仓为0的标的。

    df, msg = tapi.query_account()
    print(msg)
    assert check_err_msg(msg)
    print(df)
    
    # 查询当前策略帐号的所有持仓
    #
    # 和 query_portfolio接口不一样。如果莫个期货合约 Long, Short两个方向都有持仓，这里是返回两条记录
    # 返回的 size 不带方向，全部为 正
    df, msg = tapi.query_position()
    assert check_err_msg(msg)
    print(df)
    
    # Query Universe
    df_univ, msg = tapi.query_universe()

    # 查询Portfolio
    #
    # 返回当前的策略帐号的Universe中所有标的的净持仓，包括持仓为0的标的。

    df_portfolio, msg = tapi.query_portfolio()
    assert check_err_msg(msg)
    assert len(df_univ) == len(df_portfolio)

    # 下单接口
    #  (task_id, msg) = place_order(code, action, price, size )
    #   action:  Buy, Short, Cover, Sell, CoverToday, CoverYesterday, SellToday, SellYesterday
    # 返回 task_id 可以用改 task_id
    task_id, msg = tapi.place_order("000718.SZ", "Buy", 4.59, 100)
    if msg.endswith('market has closed'):
        pass
        task_id = -1
    else:
        assert check_err_msg(msg)
    print("task_id:", task_id)
    
    df_order, msg = tapi.query_order(task_id=task_id)
    assert check_err_msg(msg)
    print(df_order)
    
    df_trade, msg = tapi.query_trade(task_id=task_id)
    assert check_err_msg(msg)
    print(df_trade)
    
    # 批量下单1：place_batch_order
    #
    # 返回task_id, msg。
    orders = [ 
        {"security":"600030.SH", "action" : "Buy", "price": 16, "size":1000},
        {"security":"600519.SH", "action" : "Buy", "price": 320, "size":1000},
        ]

    task_id, msg = tapi.place_batch_order(orders, "", dict())
    print(task_id)
    print(msg)    

    # cancel_order
    # 撤单
    tapi.cancel_order(task_id)

    # 批量下单2：basket_order
    #
    # 返回task_id, msg。
    orders = [ 
        {"security":"601857.SH", "ref_price": 8.40, "inc_size":1000},
        {"security":"601997.SH",  "ref_price": 14.540, "inc_size":20000},
        ]

    task_id, msg = tapi.basket_order(orders, "", {})
    print(task_id)
    print(msg)

    #  goal_protfolio
    #  参数：目标持仓
    #  返回：(result, msg)
    #     result:  成功或失败
    #     msg:     错误原因
    #  注意：目标持仓中必须包括所有的代码的持仓，即使不修改
    
    # 先查询当前的持仓, 
    portfolio, msg = tapi.query_portfolio()
    print("msg", msg)
    print("portfolio", portfolio)
    
    goal = pd.DataFrame(portfolio['current_size'])
    goal.loc[:, 'size'] = goal['current_size']
    goal.loc[:, 'ref_price'] = 0.0
    goal.loc[:, 'urgency'] = 5

    #  然后修改目标持仓
    code = '601857.SH'
    goal.loc[code, 'ref_price'] = 8.38
    goal.loc[code, 'size'] += 20000

    code = '601997.SH'
    goal.loc[code, 'ref_price'] = 14.40
    goal.loc[code, 'size'] += 10000

    # stop_portfolio
    # 撤单, 撤销所有portfolio订单
    tapi.stop_portfolio()

    # 发送请求
    result, msg = tapi.goal_portfolio(goal)
    print(result, msg)


def check_err_msg(err_msg):
    l = err_msg.split(',')
    return l and (l[0] == '0')
    
if __name__ == "__main__":
    test_trade_api()
